// Description: initializes a DCS from a named node within an OpenFlight file.
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> a source and node are specified
//<li> default source is called "fltSource"
//<li> default node is the name of the fltTransform
//<li> desired node transformation is given to fltTransform node
//</ul>
//
// Category: Transformation
// Author: Alex Hill
// Revision: 11/01/01
//
#include <Performer/pf/pfDCS.h>
#include "ygNetMessages.h"
#include "ygNodeDB.h"
#include "fltTransform.h"

extern "C" ygNode* construct_fltTransform(const char* name,bool master) { return new fltTransform(name,master); }

struct _fltTransformPrivateData
    {
    pfDCS * dcs;
    pfMatrix netMatrix;
    ygString sourceNode;
    ygString sourceName;
    int sourceMode;
    pfVec3 position;
    pfVec3 orientation;
    pfVec3 size;
    };


fltTransform::fltTransform(const char* name,bool master) : ygNode(name,master,false)
    {
    setClassName("fltTransform",true);
    p_ = new struct _fltTransformPrivateData;
    p_->dcs = new pfDCS;
    setPfNode(p_->dcs);
    p_->netMatrix.makeIdent();
    p_->sourceMode = 0;
    p_->sourceName = this->name();
    addNetMessage("matrix", &p_->netMatrix, YG_NET_MATRIX);
    addNetMessage("source",&p_->sourceNode,YG_NET_STRING);
    addNetMessage("source",&p_->sourceName,YG_NET_STRING);
    addNetMessage("source",&p_->sourceMode,YG_NET_INT);
    addNetMessageMap("source","none",0);
    addNetMessageMap("source","parent",1);
    addNetMessageMap("source","child",2);
    dontSendNetMessage("source");
    addNetMessage("position",&p_->position,YG_NET_VEC3);
    addNetMessage("orientation",&p_->orientation,YG_NET_VEC3);
    addNetMessage("size",&p_->size,YG_NET_VEC3);
    }


fltTransform::~fltTransform(void)
    {
    delete p_;
    }


void fltTransform::reset(void)
    {
    if (isNetMaster())
        {
        pfMatrix mat;
        mat.makeIdent();
        setMatrix(mat);
        }
    ygNode::reset();
    }


/*****  DCS functions *****/

void fltTransform::getMatrix(pfMatrix& mat) const
    {
    p_->dcs->getMat(mat);
    }


void fltTransform::setMatrix(pfMatrix& mat)
    {
    p_->dcs->setMat(mat);
    storeMatrixKey();
    }


void fltTransform::setPosition(float x,float y,float z)
    {
    p_->dcs->setTrans(x,y,z);
    storeMatrixKey();
    }


void fltTransform::setOrientation(float x,float y,float z)
    {
    p_->dcs->setRot(z,x,y);
    storeMatrixKey();
    }


void fltTransform::setSize(float s)
    {
    p_->dcs->setScale(s);
    storeMatrixKey();
    }


void fltTransform::setSize(float x,float y,float z)
    {
    p_->dcs->setScale(x,y,z);
    storeMatrixKey();
    }


/***** Message functions *****/

void fltTransform::message(const ygMessage& msg)
    {
    if (msg == "node")
            {
            ygString fltSource("fltSource");  
            if (msg.numArgs() == 0) 
                    extractTransform(fltSource,name(),0);
            else if (msg.numArgs() == 1)
                    extractTransform(fltSource,msg.stringArg(0),1);
                else if (msg.numArgs() == 2)
                        if (msg.stringArg(1) == "child")
                           extractTransform(fltSource,msg.stringArg(0),2);
                        else
                           extractTransform(fltSource,msg.stringArg(0),0);
            else
                    msg.error(name(),"(wrong number of arguments)");
            }
    else if (msg == "source")
            {
        if (msg.numArgs() == 2)
            p_->sourceMode = 1;
        extractTransform(p_->sourceNode,p_->sourceName,p_->sourceMode);
            }
    else if (msg == "position")
        {
        setPosition(p_->position);
        }
    else if (msg == "orientation")
        {
        setOrientation(p_->orientation);
        }
    else if (msg == "size")
        {
        setSize(p_->size);
        }
    else
        ygNode::message(msg);
    }


void fltTransform::extractTransform(const ygString& source,const ygString& node,const int& mode)
    {
    ygNode *ygSource = ygNodeDB::find(source.c_str());
    if (ygSource) 
        {
        pfGroup* pointer;
        pfSCS *scsNode = NULL;
        pointer = (pfGroup*) ygSource->pfnode()->find(node.c_str(), pfGroup::getClassType());
        if (mode == 0)
            scsNode = (pfSCS*) pointer;
        else if (mode == 1)
            scsNode = (pfSCS*) pointer->getParent(0);
        else if (mode == 2)
            scsNode = (pfSCS*) pointer->getChild(0);
        if (scsNode) 
            {
            pfMatrix scs;
            scsNode->getMat(scs);
            setMatrix(scs);
            p_->sourceNode = source;
            p_->sourceName = node;
            p_->sourceMode = mode;
            netMessageChanged("source");
            }
        else
        printf("fltTransform: ERROR - node %s not found\n",node.c_str());
        }
    else
        printf("fltTransform: ERROR - source node %s not found\n",source.c_str());
    }


void fltTransform::storeMatrixKey(void)
    {
    if (isNetMaster())
        {
        getMatrix(p_->netMatrix);
        pfCoord coord;
        p_->netMatrix.getOrthoCoord(&coord);
        p_->position = coord.xyz;
        p_->orientation[0] = coord.hpr[1];
        p_->orientation[1] = coord.hpr[2];
        p_->orientation[2] = coord.hpr[0];
        pfVec3 row;
        row.set(p_->netMatrix.mat[0][0],p_->netMatrix.mat[0][1],p_->netMatrix.mat[0][2]);
        p_->size[0] = row.length();
        row.set(p_->netMatrix.mat[1][0],p_->netMatrix.mat[1][1],p_->netMatrix.mat[1][2]);
        p_->size[1] = row.length();
        row.set(p_->netMatrix.mat[2][0],p_->netMatrix.mat[2][1],p_->netMatrix.mat[2][2]);
        p_->size[2] = row.length();
        netMessageChanged("matrix");
        }
    }
