// Description: adjusts the material properties of all nodes below it
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the "relative" option makes changes relative to the current material property
//<li> the material property of all instanced object nodes will be changed
//<li> use the cache option on child objects to avoid unwanted side effects
//</ul>
//
// Category: Deprecated
// Author: Alex Hill
//           11/01/01
// Revision: 04/01/04 Alex Hill - added "alpha" message to adjust transparency
//
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pf/pfChannel.h>
#include "ygNetMessages.h"
#include "materialProperty.h"

#define ABSOLUTE 0
#define RELATIVE 1

extern "C" ygNode* construct_materialProperty(const char* name,bool master) { return new materialProperty(name,master); }

materialProperty::materialProperty(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("materialProperty",true);
    gSetList = new pfList;
    ambient.set(0,0,0);
    diffuse.set(0,0,0);
    emission.set(0,0,0);
    specular.set(0,0,0);
    alpha = 1.0;
    shininess = 0.0;
    shininessMode = false;
    ambientMode = ABSOLUTE;
    diffuseMode = ABSOLUTE;
    emissionMode = ABSOLUTE;
    specularMode = ABSOLUTE;
    alphaMode = ABSOLUTE;
    shininessMode = ABSOLUTE;
    addNetMessage("ambient",&ambient,YG_NET_VEC3);
    addNetMessage("ambient",&ambientMode,YG_NET_INT);
    addNetMessageMap("ambient","absolute",ABSOLUTE);
    addNetMessageMap("ambient","rel",RELATIVE);
    addNetMessageMap("ambient","relative",RELATIVE);
    addNetMessage("diffuse",&diffuse,YG_NET_VEC3);
    addNetMessage("diffuse",&diffuseMode,YG_NET_INT);
    addNetMessageMap("diffuse","absolute",ABSOLUTE);
    addNetMessageMap("diffuse","rel",RELATIVE);
    addNetMessageMap("diffuse","relative",RELATIVE);
    addNetMessage("emission",&emission,YG_NET_VEC3);
    addNetMessage("emission",&emissionMode,YG_NET_INT);
    addNetMessageMap("emission","absolute",ABSOLUTE);
    addNetMessageMap("emission","rel",RELATIVE);
    addNetMessageMap("emission","relative",RELATIVE);
    addNetMessage("specular",&specular,YG_NET_VEC3);
    addNetMessage("specular",&specularMode,YG_NET_INT);
    addNetMessageMap("specular","absolute",ABSOLUTE);
    addNetMessageMap("specular","rel",RELATIVE);
    addNetMessageMap("specular","relative",RELATIVE);
    addNetMessage("alpha",&alpha,YG_NET_FLOAT);
    addNetMessage("alpha",&alphaMode,YG_NET_INT);
    addNetMessageMap("alpha","absolute",ABSOLUTE);
    addNetMessageMap("alpha","rel",RELATIVE);
    addNetMessageMap("alpha","relative",RELATIVE);
    addNetMessage("shininess",&shininess,YG_NET_FLOAT);
    addNetMessage("shininess",&shininessMode,YG_NET_INT);
    addNetMessageMap("shininess","absolute",ABSOLUTE);
    addNetMessageMap("shininess","rel",RELATIVE);
    addNetMessageMap("shininess","relative",RELATIVE);
    }

materialProperty::~materialProperty(void)
    {
    }

void materialProperty::reset(void)
    {
    //reset all colors and shininess to zero
    gSetList = new pfList;
    resetNetMessage("ambient");
    resetNetMessage("diffuse");
    resetNetMessage("emission");
    resetNetMessage("specular");
    resetNetMessage("alpha");
    resetNetMessage("shininess");
    ygNode::reset();
    } 

void materialProperty::message(const ygMessage& msg)
    {
    //set or change the ambient color of the material
    if (msg == "ambient")
        {
        if (ambientMode == RELATIVE)
            changeColor(GL_AMBIENT,ambient[0],ambient[1],ambient[2]);
        else
            setColor(GL_AMBIENT,ambient[0],ambient[1],ambient[2]);
        }
    //set or change the diffuse color of the material
    else if (msg == "diffuse")
        {
        if (diffuseMode == RELATIVE)
            changeColor(GL_DIFFUSE,diffuse[0],diffuse[1],diffuse[2]);
        else
            setColor(GL_DIFFUSE,diffuse[0],diffuse[1],diffuse[2]);
        }
    //set or change the emission color of the material
    else if (msg == "emission")
        {
        if (emissionMode == RELATIVE)
            changeColor(GL_EMISSION,emission[0],emission[1],emission[2]);
        else
            setColor(GL_EMISSION,emission[0],emission[1],emission[2]);
        }
    //set or change the specular color of the material
    else if (msg == "specular")
        {
        if (specularMode == RELATIVE)
            changeColor(GL_SPECULAR,specular[0],specular[1],specular[2]);
        else
            setColor(GL_SPECULAR,specular[0],specular[1],specular[2]);
        }
    //set or change the alpha of the material
    else if (msg == "alpha")
        {
        if (alphaMode == RELATIVE)
            changeAlpha(alpha);
        else
            setAlpha(alpha);
        }
    //set or change the shininess of the material
    else if (msg == "shininess")
        {
        if (shininessMode == RELATIVE)
            changeShininess(shininess);
        else
            setShininess(shininess);
        }
    else
        ygNode::message(msg);
    }
    
void materialProperty::app(void)
    {
    //if first frame then set the geoset list
    if (isChanged(YG_CHANGE_GEOMETRY))
        setGSetList(pfnode());
    ygNode::app();
    }
    
void materialProperty::setGSetList(pfNode *node)
    {
    int i;
    //if node type is a pfGeode
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        //for each geoset
        for (i=0; i < geode->getNumGSets(); i++)
            {
            //add geoset to the geoset list
            pfGeoSet* gSet = geode->getGSet(i);
            gSetList->add(gSet);
            //duplicate geometry state to prevent material property sharing
            pfGeoState* gState = gSet->getGState();
            pfGeoState* newGState = new pfGeoState;
            newGState->copy(gState);
            gSet->setGState(newGState);
            gState = newGState;
            pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
            if (frontMtl)
                {
                pfMaterial* newMtl = new pfMaterial;
                newMtl->copy(frontMtl);
                gState->setAttr(PFSTATE_FRONTMTL,newMtl);
                gState->setAttr(PFSTATE_BACKMTL,newMtl);
                frontMtl = newMtl;
                }
            pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
            if (backMtl && backMtl != frontMtl)
                {
                pfMaterial* newMtl = new pfMaterial;
                newMtl->copy(backMtl);
                gState->setAttr(PFSTATE_BACKMTL,newMtl);
                }
            }
        }
    //else, if node type is a pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call setGSetList recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setGSetList(group->getChild(i));
        }
    }

void materialProperty::setColor(int color,float r,float g,float b)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    frontMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    frontMode = PFMTL_CMODE_DIFFUSE;
            if (color == frontMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] = r;
                    colors[i][1] = g;
                    colors[i][2] = b;
                    }
                }
            else
                frontMtl->setColor(color,r,g,b);
            }
        if (backMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    backMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    backMode = PFMTL_CMODE_DIFFUSE;
            if (backMode != frontMode && color == backMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] = r;
                    colors[i][1] = g;
                    colors[i][2] = b;
                    }
                }
            else
                backMtl->setColor(color,r,g,b);
            }
        }
    }
    
void materialProperty::changeColor(int color,float r,float g,float b)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    frontMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    frontMode = PFMTL_CMODE_DIFFUSE;
            if (color == frontMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] += r;
                    colors[i][1] += g;
                    colors[i][2] += b;
                    }
                }
            else
                {
                float currR,currG,currB;
                frontMtl->getColor(color,&currR,&currG,&currB);
                frontMtl->setColor(color,currR+r,currG+g,currB+b);
                }
            }
        if (backMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode == PFMTL_CMODE_AMBIENT_AND_DIFFUSE)
                if (color == PFMTL_AMBIENT)
                    backMode = PFMTL_CMODE_AMBIENT;
                else if (color == PFMTL_DIFFUSE)
                    backMode = PFMTL_CMODE_DIFFUSE;
            if (backMode != frontMode && color == backMode)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    {
                    colors[i][0] += r;
                    colors[i][1] += g;
                    colors[i][2] += b;
                    }
                }
            else
                {
                float currR,currG,currB;
                frontMtl->getColor(color,&currR,&currG,&currB);
                frontMtl->setColor(color,currR+r,currG+g,currB+b);
                }
            }
        }
    }

void materialProperty::setAlpha(float a)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        gSet->setDrawBin(PFSORT_TRANSP_BIN);
        pfGeoState* gState = gSet->getGState();
        gState->setMode(PFSTATE_TRANSPARENCY, PFTR_BLEND_ALPHA);
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] = a;
                }
            else
                frontMtl->setAlpha(a);
            }
        if (backMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] = a;
                }
            else
                backMtl->setAlpha(a);
            }
        }
    }
    
void materialProperty::changeAlpha(float a)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        gSet->setDrawBin(PFSORT_TRANSP_BIN);
        pfGeoState* gState = gSet->getGState();
        gState->setMode(PFSTATE_TRANSPARENCY, PFTR_BLEND_ALPHA);
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        int frontMode,backMode;
        if (frontMtl)
            {
            frontMode = frontMtl->getColorMode(PFMTL_FRONT);
            if (frontMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] += a;
                }
            else
                {
                float currA;
                currA = frontMtl->getAlpha();
                frontMtl->setAlpha(currA+a);
                }
            }
        if (backMtl)
            {
            backMode = backMtl->getColorMode(PFMTL_BACK);
            if (backMode != PFMTL_CMODE_OFF)
                {
                ushort* indexList;
                pfVec4* colors;
                gSet->getAttrLists(PFGS_COLOR4,(void **)&colors,&indexList);
                int numColors = pfGetSize(colors)/sizeof(pfVec4);
                for (int i=0;i<numColors;i++)
                    colors[i][3] += a;
                }
            else
                {
                float currA;
                currA = backMtl->getAlpha();
                backMtl->setAlpha(currA+a);
                }
            }
        }
    }

void materialProperty::setShininess(float s)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        if (frontMtl)
            frontMtl->setShininess(s);
        if (backMtl)
            backMtl->setShininess(s);
        }
    }
    
void materialProperty::changeShininess(float s)
    {
    for (int i=0;i<gSetList->getNum();i++)
        {
        pfGeoSet* gSet = (pfGeoSet*)gSetList->get(i);
        pfGeoState* gState = gSet->getGState();
        pfMaterial* frontMtl = (pfMaterial*)gState->getAttr(PFSTATE_FRONTMTL);
        pfMaterial* backMtl = (pfMaterial*)gState->getAttr(PFSTATE_BACKMTL);
        if (frontMtl)
            {
            float currS = frontMtl->getShininess();
            frontMtl->setShininess(currS+s);
            }
        if (backMtl)
            {
            float currS = backMtl->getShininess();
            backMtl->setShininess(currS+s);
            }
        }
    }
