// Description: a tri-strip surface of variable size
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the surface is created with the vertices message
//</ul>
//
// Category: Geometry
// Author: Alex Hill
// Revision: 11/01/02
//
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfGeode.h>
#include <ygPFTextureCache.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "surface.h"

using namespace std;

extern "C" ygNode* construct_surface(const char* name,bool master) { return new surface(name,master); }

surface::surface(const char* name,bool master) : ygGeometry(name,master)
    {
    setClassName("surface",true);
    color_.set(1,1,1,1);
    xSize_ = 0;
    ySize_ = 0;
    colors = NULL;
    texture = NULL;
    gstate = NULL;
    geode = NULL;
    addNetMessage("vertices",&xSize_,YG_NET_INT);
    addNetMessage("vertices",&ySize_,YG_NET_INT);
    addNetMessage("color",&color_,YG_NET_VEC4);
    addNetMessage("file",&textureFile,YG_NET_STRING);
    }

surface::~surface(void)
    {
    }

void surface::reset(void)
    {
    //reset color to opaque white
    resetNetMessage("color");
    //reset vertices to zero
    resetNetMessage("vertices");
    //reset texture file name
    resetNetMessage("file");
    colors = NULL;
    texture = NULL;
    gstate = NULL;
    //remove any existing child geode
    if (geode)
        pfnode()->removeChild(geode);
    geode = NULL;
    ygGeometry::reset();    
    }


void surface::message(const ygMessage& msg)
    {
    //determine the number of vertices
    if (msg == "vertices") 
        {
        createGeometry(xSize_,ySize_);
        }
    //set the surface color
    else if (msg == "color") 
        {
        setColor(color_);
        }
    //set the surface texture
    else if (msg == "file") 
        {
        setTexture(textureFile);
        }
    //set the surface texture (deprecated)
    else if (msg == "texture") 
        {
        if (msg.numArgs() == 1)
            setTexture(msg.stringArg(0));
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    else
        ygGeometry::message(msg);
    }

void surface::createGeometry(int xSize, int ySize)
    {
    if (geode)
        pfnode()->removeChild(geode);
    xSize_ = xSize;
    ySize_ = ySize;
    
    //allocate vertex, normal, color, and texture coordinate arrays
    verts  = (pfVec3*) pfMalloc(xSize * ySize * sizeof(pfVec3), pfGetSharedArena());
    norms  = (pfVec3*) pfMalloc(xSize * ySize * sizeof(pfVec3), pfGetSharedArena());
    colors = (pfVec4*) pfMalloc(xSize * ySize * sizeof(pfVec4), pfGetSharedArena());
    texCoords  = (pfVec2*) pfMalloc(xSize * ySize * sizeof(pfVec2), pfGetSharedArena());

    //set up vertex, normal, color, and texture corrdinate values
    int index;
    for (int y = 0, index = 0; y < ySize; y++)
        {
        for (int x = 0; x < xSize; x++, index++) 
            {
            verts[index].set(x/(float)(xSize-1) - 0.5,y/(float)(ySize-1) - 0.5,0.0);
            norms[index].set(0, 0, 1);
            colors[index].set(color_[0], color_[1], color_[2], color_[3]);
            texCoords[index].set(x/(float)(xSize-1), y/(float)(ySize-1));
            }
        }

    //create geoset
    gset = new pfGeoSet;
    gset->setPrimType(PFGS_TRISTRIPS);
    gset->setNumPrims(ySize-1);

    int* lengths = (int*)pfMalloc((ySize-1)*sizeof(int), pfGetSharedArena());
    for (int y = 0; y < ySize-1; y++)
        lengths[y] = xSize * 2;
    gset->setPrimLengths(lengths);

    //set up indexing
    ilist = (ushort*)pfMalloc(2*xSize*(ySize-1)*sizeof(ushort), pfGetSharedArena());
    index = 0;
    for (int y=0; y<ySize-1; y++)
        {
        for (int x=0; x<xSize; x++)
            {
            ilist[index++] = x + y*xSize + xSize;
            ilist[index++] = x + y*xSize;
            }
        }

    //add array attributes to geoset
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, ilist);
    gset->setAttr(PFGS_NORMAL3, PFGS_PER_VERTEX, norms, ilist);
    gset->setAttr(PFGS_COLOR4, PFGS_PER_VERTEX, colors, ilist);
    gset->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, texCoords, ilist);               

    //create geostate
    gstate = new pfGeoState;
    mat = new pfMaterial();;
    gstate->setAttr(PFSTATE_FRONTMTL, mat);
    gstate->setAttr(PFSTATE_BACKMTL, mat);
    gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
    gset->setGState(gstate);
    if (texture)
        {
        int comp;
        //get number of image components
        texture->getImage(NULL,&comp,NULL,NULL,NULL);
        //if alpha channel then set transparency
        if (comp == 4)
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PF_ON|PFTR_NO_OCCLUDE|PFTR_BLEND_ALPHA);
            gset->setDrawBin(PFSORT_TRANSP_BIN);
            }
        else
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PF_OFF);
            gset->setDrawBin(PFSORT_OPAQUE_BIN);
            }
        gstate->setAttr(PFSTATE_TEXTURE,texture);
        gstate->setMode(PFSTATE_ENTEXTURE,PF_ON);
        setChanged(YG_CHANGE_TEXTURE);
        }

    //create geode
    geode = new pfGeode;
    geode->addGSet(gset);
    pfnode()->addChild(geode);
    setChanged(YG_CHANGE_GEOMETRY);
    }

void surface::setColor(pfVec4 c)
    {
    if (colors)
        {
        color_ = c;
        for (int y = 0, index = 0; y < ySize_; y++) 
            for (int x = 0; x < xSize_; x++, index++) 
                colors[index].set(color_[0], color_[1], color_[2], color_[3]);
        }
    }

void surface::setTexture(const ygString& file)
    {
    textureFile = file;
    if (textureFile.length() > 0)
        texture = ygPFTextureCache::textureReference(file.c_str());
    else
        texture = NULL;
    if (texture && gstate)
        {
        int comp;
        //get number of image components
        texture->getImage(NULL,&comp,NULL,NULL,NULL);
        //if alpha channel then set transparency
        if (comp == 4)
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PF_ON|PFTR_NO_OCCLUDE|PFTR_BLEND_ALPHA);
            gset->setDrawBin(PFSORT_TRANSP_BIN);
            }
        else
            {
            gstate->setMode(PFSTATE_TRANSPARENCY,PF_OFF);
            gset->setDrawBin(PFSORT_OPAQUE_BIN);
            }
        gstate->setAttr(PFSTATE_TEXTURE, texture);
        gstate->setMode(PFSTATE_ENTEXTURE, PF_ON);
        setChanged(YG_CHANGE_TEXTURE);
        }
    }

