// Description: creates a light source with position, ambient, diffuse, and specular settings
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: light
//<li> this node creates a pfLightSource node
//<li> by default light comes from infinity in the direction vector formed by the given position and the origin
//<li> a fourth position argument of 1.0 creates a point light source located at the given position
//<li> lighting in a scene is based on the number of lights, lighting-model modes, and the material properties of the illuminated objects where the contribution of each light source is the sum of:
//<ol>
//<li> light source ambient color, scaled by the material ambient reflectance
//<li> light source diffuse color, scaled by the material diffuse reflectance and the dot product of the vertex normal and the vertex-to-light source vector
//<li> light source specular color, scaled by the material specular reflectance and a function of the angle between the vertex-to-viewpoint vector and the vertex-to-light source vector
//</ol>
//<li> the ambient component is nondirectional and distributed uniformly throughout the environment
//<li> the ambient component is reflected from the object independent of surface location and orientation with equal intensity in all directions
//<li> the diffuse component takes into account the direction of a light source. The intensity of light striking a surface varies with the angle between the orientation of the object and the direction of the light source -- the diffuse parameter of a material determines how much light is scattered evenly in all directions
//<li> the specular component represents reflection off shiny surfaces. Specular lighting operates based on the viewer's position. Maximum specular reflectance occurs when the angle between the viewer and the direction of the reflected light is zero -- the brightness of a specular reflection decays based upon the exponential value shininess
//</ul>
//
// Category: Attributes
// Author: Dave Pape
//           11/01/01 Geoffrey Baum - added ambient and specular messages
// Revision: 06/28/05 Alex Hill - fix position message syntax 
//
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfGroup.h>
#include "ygNetMessages.h"
#include "ygLight.h"

extern "C" ygNode* construct_ygLight(const char* name,bool master) { return new ygLight(name,master); }

struct _ygLightPrivateData
    {
    pfLightSource * light;
    struct
        {
        pfVec4 position;
        pfVec3 ambient;
        pfVec3 diffuse;
        pfVec3 specular;
        bool on;
        } net;
    bool prevOn;
    };


ygLight::ygLight(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygLight",true);
    p_ = new struct _ygLightPrivateData;
    //create pfLightSource
    p_->light = new pfLightSource;
    pfnode()->addChild(p_->light);
    p_->net.position.set(0,0,1,0);
    p_->net.ambient.set(0,0,0);
    p_->net.diffuse.set(1,1,1);
    p_->net.specular.set(1,1,1);
    p_->net.on = true;
    addNetMessage("position",&p_->net.position,YG_NET_VEC4);
    addNetMessage("ambient",&p_->net.ambient,YG_NET_VEC3);
    addNetMessage("diffuse",&p_->net.diffuse,YG_NET_VEC3);
    addNetMessage("specular",&p_->net.specular,YG_NET_VEC3);
    addNetMessage("off",&p_->net.on,YG_NET_BOOL_NEGATE);
    dontSendNetMessage("off");
    }


ygLight::~ygLight(void)
    {
    pfnode()->removeChild(p_->light);
    pfDelete(p_->light);
    delete p_;
    }


void ygLight::reset(void)
    {
    //if net master
    if (isNetMaster())
        {
        //set position to 0,0,1,0
        resetNetMessage("position");
        //set ambient to 0,0,0
        resetNetMessage("ambient");
        //set diffuse to 1,1,1
        resetNetMessage("diffuse");
        //set specular to 1,1,1
        resetNetMessage("specular");
        //set to on
        on();
        p_->prevOn = true;
        }
    ygNode::reset();
    }


void ygLight::app(void)
    {
    bool ison = isOn();
    if ((ison) && (!p_->prevOn))
        eventOccurred("LightOn");
    else if ((!ison) && (p_->prevOn))
        eventOccurred("LightOff");
    p_->prevOn = ison;
    ygNode::app();
    }



void ygLight::message(const ygMessage& msg)
    {
    //set the light to on
    if (msg == "on")
        {
        on();
        }
    //set the light to off
    else if (msg == "off")
        {
        if (p_->net.on)
           on();
        else
           off();
        }
    //toggle the light state
    else if (msg == "toggle")
        {
        toggle();
        }
    //set the position from which the light comes
    else if (msg == "position")
        {
        if (msg.numArgs() == 3)
            {
            pfVec3 vec;
            msg.getVec3Args(vec);
            p_->net.position[0] = vec[0];
            p_->net.position[1] = vec[1];
            p_->net.position[2] = vec[2];
            p_->net.position[3] = 0.0;
            }
        else if (msg.numArgs() == 4)
            {
            msg.getVec4Args(p_->net.position);
            }
        p_->light->setPos(p_->net.position[0],p_->net.position[1],p_->net.position[2],p_->net.position[3]);
        }
    //set the ambient color of the light
    else if (msg == "ambient")
        {
        p_->light->setColor(PFLT_AMBIENT,p_->net.ambient[0],p_->net.ambient[1],p_->net.ambient[2]);
        }
    //set the diffuse color of the light
    else if (msg == "diffuse")
        {
        p_->light->setColor(PFLT_DIFFUSE,p_->net.diffuse[0],p_->net.diffuse[1],p_->net.diffuse[2]);
        }
    //set the specular color of the light
    else if (msg == "specular")
        {
        p_->light->setColor(PFLT_SPECULAR,p_->net.specular[0],p_->net.specular[1],p_->net.specular[2
]);
        }
    else
        ygNode::message(msg);
    }


/*****  LightSource functions *****/

bool ygLight::isOn(void) const
    {
    return (p_->light->isOn() != 0);
    }

void ygLight::on(void)
    {
    if (!isOn())
        {
        p_->light->on();
        p_->net.on = true;
        netMessageChanged("off");
        }
    }

void ygLight::off(void)
    {
    if (isOn())
        {
        p_->light->off();
        p_->net.on = false;
        netMessageChanged("off");
        }
    }

void ygLight::toggle(void)
    {
    if (isOn())
        off();
    else
        on();
    }

void ygLight::setPosition(float x,float y,float z,float w)
    {
    p_->light->setPos(x,y,z,w);
    p_->net.position.set(x,y,z,w);
    netMessageChanged("position");
    }


void ygLight::setAmbient(float r,float g,float b)
    {
    p_->light->setColor(PFLT_AMBIENT, r, g, b);
    p_->net.ambient.set(r,g,b);
    netMessageChanged("ambient");
    }

void ygLight::setDiffuse(float r,float g,float b)
    {
    p_->light->setColor(PFLT_DIFFUSE, r, g, b);
    p_->net.diffuse.set(r,g,b);
    netMessageChanged("diffuse");
    }

void ygLight::setSpecular(float r,float g,float b)
    {
    p_->light->setColor(PFLT_SPECULAR, r, g, b);
    p_->net.specular.set(r,g,b);
    netMessageChanged("specular");
    }
