// Description: creates a transformation matrix with position, orientation, and scale
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the aliases: transform, ygSimpleTransform, and simpleTransform
//<li> the orientation message is measured in degrees and allows you to control the heading, pitch, and roll of the child node
//<li> creates a pfDCS and maintains network state with a pfCoord for translation and rotation and a pfVec3 for size
//</ul>
//
// Category: Transformation
// Author: Dave Pape
//           11/01/01
//           09/01/04 Alex Hill - merged with ygSimpleTransform and added "coordinates" debug
// Revision: 06/28/05 Alex Hill - fixed orientation update problem in setMatrix
//
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pfdu.h>
#include "ygNetMessages.h"
#include "ygWorld.h"
#include "ygTransform.h"
#include "ygUtil.h"

extern "C" ygNode* construct_ygTransform(const char* name,bool master) { return new ygTransform(name,master); }

#define COORD_NETKEY "x"
#define SIZE_NETKEY "s"
#define MATRIX_NETKEY "m"

struct _ygTransformPrivateData
    {
    pfDCS* dcs;
    pfCoord netCoord;
    pfCoord lastCoord;
    pfVec3 netSize;
    pfVec3 lastSize;
    pfVec3 pos;
    pfVec3 ori;
#if 0
    pfMatrix netMatrix;
#endif
    float updateInterval, lastUpdate;
    bool debugCoordinates;
    pfGeode* debugGeode;
    };


ygTransform::ygTransform(const char* name,bool master) : ygNode(name,master,false)
    {
    setClassName("ygTransform",true);
    p_ = new struct _ygTransformPrivateData;
    //create new pfDCS and set as pfnode
    p_->dcs = new pfDCS;
    setPfNode(p_->dcs);
    p_->netCoord.xyz.set(0,0,0);
    p_->netCoord.hpr.set(0,0,0);
    p_->lastCoord = p_->netCoord;
    p_->netSize.set(1,1,1);
    p_->ori.set(0,0,0);
    p_->netSize.set(1,1,1);
    p_->lastSize = p_->netSize;
    p_->updateInterval = 0;
    p_->lastUpdate = 0;
    p_->debugGeode = NULL;
    //show a wireframe of the coordinates
    debugFlag("coordinates", &p_->debugCoordinates);
    addNetMessage("size",&p_->netSize,YG_NET_VEC3);
    dontSendNetMessage("size");
    addNetMessage("position",&p_->netCoord.xyz,YG_NET_VEC3);
    dontSendNetMessage("position");
    addNetMessage("orientation",&p_->ori,YG_NET_VEC3);
    dontSendNetMessage("orientation");
    addNetMessage("updateInterval",&p_->updateInterval,YG_NET_FLOAT);
    //create a simple unreliable coordinate network key
    addNetMessage(COORD_NETKEY, &p_->netCoord, YG_NET_COORD);
    //dont include in documentation
    dontSaveNetMessage(COORD_NETKEY);
    unreliableMessage(COORD_NETKEY);
#if 0
    addNetMessage(MATRIX_NETKEY, &p_->netMatrix, YG_NET_MATRIX);
    //dont include in documentation
    dontSaveNetMessage(MATRIX_NETKEY);
    unreliableMessage(MATRIX_NETKEY);
#endif
    }


ygTransform::~ygTransform(void)
    {
    delete p_;
    }


void ygTransform::reset(void)
    {
    //if net master then set matrix to identity
    if (isNetMaster())
        {
        pfMatrix mat;
        mat.makeIdent();
        setMatrix(mat);
        //reset update interval
        resetNetMessage("updateInterval");
        //reset last update timer
        p_->lastUpdate = 0;
        }
    ygNode::reset();
    }


void ygTransform::message(const ygMessage& msg)
    {
    //set the translational component
    if (msg == "position")
        {
        setPosition(p_->netCoord.xyz);
        }
    //set the rotational component applied before translation
    else if (msg == "orientation")
        {
        setOrientation(p_->ori);
        }    
    //set the scaling component
    else if (msg == "size")
        {
        if (msg.numArgs() == 1)
            {
            float val;
            val = msg.floatArg(0);
            p_->netSize[0] = val;
            p_->netSize[1] = val;
            p_->netSize[2] = val;
            }
        else if (msg.numArgs() == 3)
            {
            msg.getVec3Args(p_->netSize);
            }
        setSize(p_->netSize);
        }
    //set the key update interval for remote users
    else if (msg == "updateInterval")
        {    
        return;
        }
    else if (msg == COORD_NETKEY)
        {
        setPosition(p_->netCoord.xyz);
        setOrientation(p_->netCoord.hpr[1],p_->netCoord.hpr[2],p_->netCoord.hpr[0]);
        }
#if 0
    else if (msg == MATRIX_NETKEY)
        {
        setMatrix(p_->netMatrix);
        }
#endif
    else
        ygNode::message(msg);
    }

void ygTransform::app(void)
    {
    //if net master
    if (isNetMaster())
        {
        //if update interval has passed then update key
        if ((ygWorld::frameTime() - p_->lastUpdate) > p_->updateInterval)
            {
            if (p_->netSize != p_->lastSize)
                {
                netMessageChanged("size");
                p_->lastSize = p_->netSize;
                p_->lastUpdate = ygWorld::frameTime();
                }
            if ((p_->netCoord.xyz != p_->lastCoord.xyz) ||
                (p_->netCoord.hpr != p_->lastCoord.hpr))
                {
                netMessageChanged(COORD_NETKEY);
                p_->lastCoord = p_->netCoord;
                p_->lastUpdate = ygWorld::frameTime();
                }
            }
        }
    ygNode::app();
    }


void ygTransform::getMatrix(pfMatrix& mat) const
    {
    p_->dcs->getMat(mat);
    }


void ygTransform::setMatrix(pfMatrix& mat)
    {
    p_->dcs->setMat(mat);
    pfMatrix matrix(mat);
    pfVec3d row;
    row.set(mat.mat[0][0],mat.mat[0][1],mat.mat[0][2]);
    p_->netSize[0] = row.length();
    row.set(mat.mat[1][0],mat.mat[1][1],mat.mat[1][2]);
    p_->netSize[1] = row.length();
    row.set(mat.mat[2][0],mat.mat[2][1],mat.mat[2][2]);
    p_->netSize[2] = row.length();
    pfMatrix unscaled;
    unscaled.preScale(1.0/p_->netSize[0],1.0/p_->netSize[1],1.0/p_->netSize[2],mat);
    unscaled.getOrthoCoord(&(p_->netCoord));
    if (p_->netCoord.hpr[0] == -180.0 && p_->netCoord.hpr[2] == 180.0)
        {
        p_->netCoord.hpr[0] = 0.0;
        if (p_->netCoord.hpr[1] >= 0.0)
            p_->netCoord.hpr[1] = 180.0 - p_->netCoord.hpr[1];
        else
            p_->netCoord.hpr[1] = -(180.0 + p_->netCoord.hpr[1]);
        p_->netCoord.hpr[2] = 0.0;
        }
    p_->ori[0] = p_->netCoord.hpr[1];
    p_->ori[1] = p_->netCoord.hpr[2];
    p_->ori[2] = p_->netCoord.hpr[0];
    }


void ygTransform::setPosition(float x,float y,float z)
    {
    p_->dcs->setTrans(x,y,z);
    p_->netCoord.xyz.set(x,y,z);
    }


void ygTransform::setOrientation(float x,float y,float z)
    {
    p_->ori.set(x,y,z);
    p_->dcs->setRot(z,x,y);
    p_->netCoord.hpr.set(z,x,y);
    }


void ygTransform::setSize(float s)
    {
    p_->netSize.set(s,s,s);
    p_->dcs->setScale(s);
    }


void ygTransform::setSize(float x,float y,float z)
    {
    p_->netSize.set(x,y,z);
    p_->dcs->setScale(x,y,z);
    }


void ygTransform::setUpdateInterval(float v)
    {
    p_->updateInterval = v;
    }


void ygTransform::updateDebug(void)
    {
    if (!p_->debugGeode)
        makeDebugGeode();
    if (p_->debugCoordinates)
        p_->debugGeode->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    else
        p_->debugGeode->setTravMask(PFTRAV_DRAW, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    }


void ygTransform::makeDebugGeode(void)
    {
    p_->debugGeode = new pfGeode;
    pfnode()->addChild(p_->debugGeode);
    if (!p_->debugGeode)
        {
        cerr << "ERROR: ygTransform::makeDebugGeode: failed to create debugGeode\n";
        return;
        }
    pfGeoSet *gset;
    pfVec4 *colorAttr;
    pfMatrix matrix;
    //create X coordinate arrow
    gset = pfdNewCone(16,pfGetSharedArena());
    colorAttr = (pfVec4 *) pfMalloc(sizeof(pfVec4),pfGetSharedArena());
    colorAttr->set(1,0,0,1);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,colorAttr,NULL);
    matrix.makeScale(0.1,0.1,0.1);
    matrix.postRot(matrix,90.0,0,1,0);
    matrix.postTrans(matrix,0.9,0,0);
    pfdXformGSet(gset,matrix);
    p_->debugGeode->addGSet(gset);
    //create Y coordinate arrow
    gset = pfdNewCone(16,pfGetSharedArena());
    colorAttr = (pfVec4 *) pfMalloc(sizeof(pfVec4),pfGetSharedArena());
    colorAttr->set(0,1,0,1);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,colorAttr,NULL);
    matrix.makeScale(0.1,0.1,0.1);
    matrix.postRot(matrix,-90.0,1,0,0);
    matrix.postTrans(matrix,0,0.9,0);
    pfdXformGSet(gset,matrix);
    p_->debugGeode->addGSet(gset);
    //create Z coordinate arrow
    gset = pfdNewCone(16,pfGetSharedArena());
    colorAttr = (pfVec4 *) pfMalloc(sizeof(pfVec4),pfGetSharedArena());
    colorAttr->set(0,0,1,1);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,colorAttr,NULL);
    matrix.makeScale(0.1,0.1,0.1);
    matrix.postTrans(matrix,0,0,0.9);
    pfdXformGSet(gset,matrix);
    p_->debugGeode->addGSet(gset);
    //create coordinate axis
    gset = new pfGeoSet;
    pfVec3 * verts = (pfVec3 *) pfMalloc(4 * sizeof(pfVec3),pfGetSharedArena());
    int *lengths = (int *) pfMalloc(3 * sizeof(int),pfGetSharedArena());
    ushort *ilist_v = (ushort *) pfMalloc(6 * sizeof(ushort),pfGetSharedArena());
    pfVec4 *colors = (pfVec4 *) pfMalloc(3 * sizeof(pfVec4),pfGetSharedArena());
    ushort *ilist_c = (ushort *) pfMalloc(3 * sizeof(ushort),pfGetSharedArena());
    gset->setPrimType(PFGS_LINES);
    gset->setNumPrims(3);
    lengths[0] = 2;
    lengths[1] = 2;
    lengths[2] = 2;
    gset->setPrimLengths(lengths);
    verts[0].set(0,0,0);
    verts[1].set(1,0,0);
    verts[2].set(0,1,0);
    verts[3].set(0,0,1);
    ilist_v[0] = 0;
    ilist_v[1] = 1;
    ilist_v[2] = 0;
    ilist_v[3] = 2;
    ilist_v[4] = 0;
    ilist_v[5] = 3;
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, ilist_v);
    colors[0].set(1,0,0,1);
    colors[1].set(0,1,0,1);
    colors[2].set(0,0,1,1);
    ilist_c[0] = 0;
    ilist_c[1] = 1;
    ilist_c[2] = 2;
    gset->setAttr(PFGS_COLOR4, PFGS_PER_PRIM, colors, ilist_c);
    gset->setLineWidth(1.0);
    p_->debugGeode->addGSet(gset);
    p_->debugGeode->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    }

