// Description: the class that loads scene files and maintains the scene graph
//
//<b>notes:</b>
//<ul>
//<li> this is not a node that can be created within Ygdrasil scripts
//<li> messages can be sent to the ygWorld instance by addressing them to "world" (i.e. world.quit)
//</ul>
//
// Category: Foundation
// Author: Dave Pape
//           11/01/01
//           10/01/02 Alex Hill - added pfChanged and setPFChanged
//           08/01/03 Alex Hill - changed ygSound references to use ygSoundServer
//           09/01/04 Alex Hill - added create, delete, and debug messages
// Revision: 06/28/05 Alex Hill - adjusted timing calculation to compenstate for long load times
//
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <list>
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pr/pfFog.h>
#include <Performer/pf/pfScene.h>
#include <pfcave.h>
#include "ygParser.h"
#include "ygUtil.h"
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygDebugFlags.h"
#include "ygSoundServer.h"
#include "ygCAVEViewer.h"
#include "ygWorld.h"

using namespace std;


class ygDelayMsg
    {
    public:
     ygMessage msg;
     float time;
     friend inline bool operator< (const ygDelayMsg& a,const ygDelayMsg& b)
            { return a.time < b.time; };
    };


struct _ygWorldPrivateData
    {
    bool sentMessageDebug;
    list<ygDelayMsg*> messageQueue;
    pfScene* pfscene;
    pfEarthSky* pfsky;
    pfFog* pffog;
    vector<ygUser*> users;
    ygNode* viewer;
    float lastTime;
    bool quitFlag;
    };



float ygWorld::frameTime_ = 0;
float ygWorld::frameDeltaTime_ = 1.0;
int ygWorld::frameNumber_ = 0;
ygWorld* ygWorld::world_ = NULL;


ygWorld::ygWorld(void)
    {
    if (world_)
        cerr << "ERROR: More than one ygWorld has been created -"
            " there should be only one!\n";
    world_ = this;
    p_ = new struct _ygWorldPrivateData;
    p_->pfscene = new pfScene;
    p_->pfsky = new pfEarthSky;
    p_->pffog = new pfFog;
    pfCAVEMasterChan()->setESky(p_->pfsky);
    p_->sentMessageDebug = ygDebugFlags::checkDebugEnv("world.sentmessages");
    ygString viewerName;
    char str[256];
    if (gethostname(str,sizeof(str)) == 0)
        viewerName = str;
    else
        viewerName = "ygdrasil";
    ygMakeValidNodeName(viewerName);
    viewerName += "_";
    sprintf(str,"%x",getpid()%1000);
    viewerName += str;
    ygString clientName;
    clientName = viewerName;
    viewerName += "_viewer";
    ygNodeDB::createValidName("ygCAVEViewer",viewerName);
    p_->viewer = new ygCAVEViewer(viewerName.c_str(),true);
    ygNodeDB::createValidName("group",clientName);
    localRoot_ = ygNodeDB::create("group",clientName);
    ygSoundServer::initServer();
    frameTime_ = ygGetTime();
    p_->quitFlag = false;
    }


ygWorld::~ygWorld(void)
    {
    root_->removeChild(localRoot_);
    ygNet::process();
    deleteLocalScene();
    while (!p_->messageQueue.empty())
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        p_->messageQueue.pop_front();
        delete dm;
        }
    delete p_;
    ygSoundServer::closeServer();
    }


void ygWorld::createRoot(void)
    {
    root_ = ygNodeDB::create("group","root");
    root_->addChild(localRoot_);
    p_->pfscene->addChild(root_->pfnode());
    }


void ygWorld::setRoot(ygNode *r)
    {
    root_ = r;
    if (root_)
        {
        root_->addChild(localRoot_);
        p_->pfscene->addChild(root_->pfnode());
        }
    }


pfScene* ygWorld::pfscene(void) const
    {
    return p_->pfscene;
    }


ygNode* ygWorld::viewer(void) const
    {
    return p_->viewer;
    }


void ygWorld::reset(void)
    {
    frameTime_ = ygGetTime();
    frameNumber_ = 0;
    while (!p_->messageQueue.empty())
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        p_->messageQueue.pop_front();
        delete dm;
        }
    if (localRoot_)
        localRoot_->resetTree();
    if (p_->viewer)
        p_->viewer->reset();
    }


void ygWorld::frame(void)
    {
    float t;
    t = ygGetTime();
    if (frameNumber_ <= 1)
        frameDeltaTime_ = 0.015;
    else
        frameDeltaTime_ = t - p_->lastTime;
    p_->lastTime = t;
    frameTime_ += frameDeltaTime_;
    if (frameDeltaTime_ <= 0)
        {
        cerr << "ERROR (ygWorld::app): Bogus delta time ("
            << frameDeltaTime_ << "); kludging to .015\n";
        frameDeltaTime_ = 0.015;
        }
    frameNumber_++;
    while ((!p_->messageQueue.empty()) &&
        (p_->messageQueue.front()->time <= frameTime_))
        {
        ygDelayMsg *dm = p_->messageQueue.front();
        sendMessage(dm->msg);
        p_->messageQueue.pop_front();
        delete dm;
        }
    if (root_)
        root_->appTraverse();
    if (p_->viewer)
        p_->viewer->app();
    ygSoundServer::updateServer();
    ygNet::process();
    }


void ygWorld::scheduleMessage(ygMessage& msg,bool queue)
    {
    if (msg.delay() <= 0 && !queue)
        sendMessage(msg);
    else
        {
        ygDelayMsg * dm = new ygDelayMsg;
        dm->time = frameTime_ + msg.delay();
        dm->msg = msg;
        list<ygDelayMsg*>::iterator i = p_->messageQueue.begin();
        for ( ; i != p_->messageQueue.end(); ++i)
            if (dm->time < (*i)->time)
                break;
        p_->messageQueue.insert(i,dm);
        }
    }


void ygWorld::sendMessage(ygMessage& msg)
    {
    if (p_->sentMessageDebug)
        cout << "ygWorld: sending message " << msg << endl;
    if (msg.recipientName() == "world")
        {
        message(msg);
        return;
        }
    if (!msg.recipient())
        msg.setRecipient(ygNodeDB::find(msg.recipientName(),false));
    if (msg.recipient())
        {
        if (msg.recipient()->isNetMaster())
            msg.recipient()->sendMessage(msg);
        else
            msg.recipient()->sendNetMessage(msg);
        }
    else
        cerr << "ERROR: could not find recipient node for message '" << msg << "'\n";
    }


void ygWorld::message(const ygMessage& msg)
    {
    //reset all local nodes
    if (msg == "reset")
        {
        reset();
        }
    //quit the application
    else if (msg == "quit" || msg == "exit")
        {
        p_->quitFlag = true;
        }
    //graph the scene from optional node and to optional depth
    else if (msg == "graph")
        {
        ygNode* node;
        int depth;
        if (msg.numArgs() > 0)
            {
            node = ygNodeDB::find(msg.stringArg(0),false);
            if (msg.numArgs() > 1)
                depth = msg.intArg(1);
            else
                depth = 99;
            }
        else
            node = ygWorld::world()->root();
        if (node)
            ygPrintSceneGraph(node,depth);
        if (!node)
            msg.error("World","(node not found)");
        }
    //load the given filename at optional node
    else if (msg == "load")
        {
        if (msg.numArgs() > 0)
            {
            ygString filename;
            filename = msg.stringArg(0);
            if (msg.numArgs() > 1)
                {
                ygNode* node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    loadFromFile(filename,node);
               else
                    msg.error("World","(node not found)");
                }
            else
                loadFromFile(filename,localRoot_);
            }
        else
            msg.error("World","(wrong number of arguments)");
        }
    //save the scene to given filename from optional node down
    else if (msg == "save")
        {
        if (msg.numArgs() > 0)
            {
            ygString filename;
            filename = msg.stringArg(0);
            ygNode* node;
            if (msg.numArgs() > 1)
                {
                node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    saveToFile(filename,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                saveLocalScene(filename);
            }
        else
            msg.error("World","(wrong number of arguments)");
        }
    //create a node with with optional name
    else if (msg == "create")
        {
        ygNode* node = NULL;
        if (msg.numArgs() > 0)
            {
            ygString name;
            ygString type;
            type = msg.stringArg(0);
            if (msg.numArgs() > 1)
                node = ygNodeDB::create(type,msg.stringArg(1));
            else
                {
                ygNodeDB::createValidName(type,name);
                node = ygNodeDB::create(type,name);
                }
            }
        else
            msg.error("World","(wrong number of arguments)");
        if (node)
            {
            localRoot()->addChild(node);
            cout << "ygWorld: created node " << node->name() << endl;
            }
        }
    //delete the given node
    else if (msg == "delete")
        {
        if (msg.numArgs() == 1)
            {
            ygNode * node = ygNodeDB::find(msg.stringArg(0),false);
            if (node)
                deleteTree(node);
            else
                msg.error("World","(node not found)");
            }
        else
            deleteLocalScene();
        }
    //set the debug state of a node type from optional node down
    else if (msg == "debug")
        {
        if (msg.numArgs() > 0)
            {
            ygString *type;
            int i=0;
            ygString message = msg.stringArg(0);
            type = message.nextToken(".",&i);
            ygString *flag;
            flag = message.nextToken(".",&i);
            ygString msgStr = "debug(";
            msgStr += *flag;
            msgStr += ")";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            if (msg.numArgs() > 1)
                {
                ygNode* node;
                node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    sendClassMessage(dbgMsg,*type,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                sendClassMessage(dbgMsg,*type,localRoot_);
            delete type;
            delete flag;
            }
        }
    //unset the debug state of a node type from optional node down
    else if (msg == "undebug")
        {
        if (msg.numArgs() > 0)
            {
            ygString *type;
            int i=0;
            ygString message = msg.stringArg(0);
            type = message.nextToken(".",&i);
            ygString *flag;
            flag = message.nextToken(".",&i);
            ygString msgStr = "debug(";
            msgStr += *flag;
            msgStr += ",false";
            msgStr += ")";
            ygMessage dbgMsg;
            dbgMsg.parseString(msgStr);
            if (msg.numArgs() > 1)
                {
                ygNode* node;
                node = ygNodeDB::find(msg.stringArg(1),false);
                if (node)
                    sendClassMessage(dbgMsg,*type,node);
                else
                    msg.error("World","(node not found)");
                }
            else
                sendClassMessage(dbgMsg,*type,localRoot_);
            delete type;
            delete flag;
            }
        else
            msg.error("World","(wrong number of arguments)");
        }
    //assume the viewpoint of another user
    else if (msg == "viewer")
        {
        if (msg.numArgs() > 0)
            {
            ygNode* node;
            node = ygNodeDB::find(msg.stringArg(0));
            if (node && node->isOfClass("ygUser"))
                {
                if (p_->viewer)
                    ((ygCAVEViewer*)p_->viewer)->setUser((ygUser*)node);
                }
            else
                msg.error("World","(user not found)");
            }
        else
            {
            if (p_->viewer)
                ((ygCAVEViewer*)p_->viewer)->setUser();
            }
        }
    else
        cout << "WARNING: " << "World received unknown message " << msg << endl;
    }


void ygWorld::addUser(ygUser* user)
    {
    p_->users.push_back(user);
    }


void ygWorld::removeUser(ygUser* user)
    {
    vector<ygUser*>::iterator i = p_->users.begin();
    while ((i != p_->users.end()) && (*i != user))
        ++i;
    if (i != p_->users.end())
        p_->users.erase(i);
    }


int ygWorld::numUsers(void) const
    {
    return p_->users.size();
    }


ygUser* ygWorld::user(int index) const
    {
    if ((index >= 0) && (index < p_->users.size()))
        return p_->users[index];
    else
        return NULL;
    }


void ygWorld::setClip(float near,float far)
    {
    CAVENear = near;
    CAVEFar = far;
    }


void ygWorld::setSkyColor(const pfVec3& sky)
    {
    p_->pfsky->setColor(PFES_CLEAR, sky[0], sky[1], sky[2], 0.0);
    }


void ygWorld::fogOff(void)
    {
    p_->pfsky->setFog(PFES_GENERAL, NULL);
    }


void ygWorld::setFog(int fogtype,const pfVec3& color,float onset,float opaque)
    {
    p_->pffog->setFogType(fogtype);
    p_->pffog->setColor(color[0],color[1],color[2]);
    p_->pffog->setRange(onset,opaque);
    p_->pfsky->setFog(PFES_GENERAL, p_->pffog);
    }


void ygWorld::deleteLocalScene(void)
    {
    while (localRoot_->numChildren() > 0)
        deleteTree(localRoot_->child(0));
    }


void ygWorld::deleteTree(ygNode* node)
    {
    if (!node)
        return;
    while (node->numChildren() > 0)
        deleteTree(node->child(0));
    if (node->parent())
        node->parent()->removeChild(node);
    delete node;
    }


void ygWorld::saveToFile(const ygString& filename,ygNode* node)
    {
    ofstream fout;
    fout.open(filename.c_str());
    saveTree(fout,node,1);
    fout.close();
    }


void ygWorld::saveLocalScene(const ygString& filename)
    {
    ofstream fout;
    fout.open(filename.c_str());
    for (int i=0; i < localRoot_->numChildren(); i++)
        saveTree(fout,localRoot_->child(i),1);
    fout.close();
    }


void ygWorld::loadFromFile(const ygString& filename,ygNode* node)
    {
    ygParser parser;
    parser.parseFile(filename,node);
    if (p_->viewer)
        p_->viewer->reset();
    }


void ygWorld::saveTree(ostream& s,ygNode* node, int depth)
    {
    if (node)
        {
        int i;
        node->printMessages(s,depth);
        depth++;
        if (node->numChildren() && node->className() != "ygScene")
            {
            for (i=0; i < depth; i++)
                s << "    ";
            s << "{" << endl;
            for (i=0; i < node->numChildren(); i++)
                saveTree(s,node->child(i),depth);
            for (i=0; i < depth; i++)
                s << "    ";
            s << "}" << endl;
            }
        }
    }


void ygWorld::sendClassMessage(const ygMessage& msg,const ygString& type,ygNode* node)
    {
    if (node)
        {
        if (type == "*" || node->isOfClass(type))
            node->sendMessage(msg);
        for (int i=0; i < node->numChildren(); i++)
            sendClassMessage(msg,type,node->child(i));
        }
    }


bool ygWorld::quit(void)
    {
    return p_->quitFlag;
    }
